/*
 * yasapi_ring1.c
 *
 * Yet Another (WA)SAPI (YASAPI) New Technology (NT) Output (out_yasapi-nt)
 * Copyright (C) 2015-2018  Peter Belkner (pbelkner@snafu.de)
 *
 * out_yasapi-nt is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * out_yasapi-nt is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with out_yasapi-nt.  If not, see <https://www.gnu.org/licenses/>.
 */
#include <yasapi.h>
#if defined (YASAPI_RING1) // [
#include <math.h>

int RingCreate(Ring *pRing, Engine *pEngine, DWORD dwSizeSource)
{
  WORD uFrames=dwSizeSource/pEngine->wfx.source.Format.nBlockAlign;
  DWORD dwSize=dwSizeSource;
  pRing->pEngine=pEngine;

  switch (pEngine->options.common.eDither) {
  case eDitherRound:
    pRing->dither.lpVtbl=&gcIRingDitherRoundVtbl;
    break;
  case eDitherSimple:
    pRing->dither.lpVtbl=&gcIRingDitherSimpleVtbl;
    break;
  case eDitherTruncate:
  default:
    pRing->dither.lpVtbl=&gcIRingDitherNullVtbl;
    break;
  }

  if (!(pRing->_buf=malloc(dwSize))) {
    DMESSAGE("allocation ring");
    goto emalloc;
  }

  pRing->_rp=pRing->_buf;
  pRing->_wp=pRing->_buf;
  pRing->_mp=pRing->_buf+dwSize;

  pRing->frames._uWritten=0u;
  pRing->frames._uAvailable=uFrames;
  DVWRITELN("uAvailable=%u uWritten=%u",pRing->frames._uAvailable,
      pRing->frames._uWritten);
emalloc:
  return 0;
}

void RingDestroy(Ring *pRing)
{
  if (pRing->_buf) {
    free(pRing->_buf);
    pRing->_buf=NULL;
  }

  pRing->pEngine=NULL;
}

void RingReset(Ring *pRing)
{
  WORD nBlockAlign=pRing->pEngine->wfx.source.Format.nBlockAlign;

  if (!pRing->pEngine) {
    DMESSAGE("ring not initialized");
    goto enoinit;
  }

  EnterCriticalSection(&pRing->pEngine->cs); // {
  pRing->_rp=pRing->_buf;
  pRing->_wp=pRing->_buf;

  pRing->frames._uWritten=0u;
  pRing->frames._uAvailable=(pRing->_mp-pRing->_buf)/nBlockAlign;
  LeaveCriticalSection(&pRing->pEngine->cs); // }
enoinit:
  return;
}

int RingWrite(Ring *pRing, const BYTE *buf, DWORD dwSize)
{
  int code=-1;
  WORD nBlockAlign=pRing->pEngine->wfx.source.Format.nBlockAlign;
	WORD uFrames=dwSize/nBlockAlign;
  DWORD dwSize1,dwSize2;

  if (!pRing->pEngine) {
    DMESSAGE("ring not initialized");
    goto enoinit;
  }

  EnterCriticalSection(&pRing->pEngine->cs); // {

  if (pRing->_wp<pRing->_rp) {
    // already wrapped
    if (pRing->_rp<pRing->_wp+dwSize) {
      DMESSAGE("overflow");
      goto eoverflow;
    }

    memcpy(pRing->_wp,buf,dwSize);
    pRing->_wp+=dwSize;
  }
  else if (pRing->_mp<pRing->_wp+dwSize) {
    // will wrap
    dwSize1=pRing->_mp-pRing->_wp;
    dwSize2=dwSize-dwSize1;

    if (pRing->_rp<pRing->_buf+dwSize2) {
      DMESSAGE("overflow");
      goto eoverflow;
    }

    memcpy(pRing->_wp,buf,dwSize1);
    memcpy(pRing->_buf,buf+dwSize1,dwSize2);
    pRing->_wp=pRing->_buf+dwSize2;
  }
  else {
    // will not wrap
    if (pRing->_mp<pRing->_wp+dwSize) {
      DMESSAGE("overflow");
      goto eoverflow;
    }

    memcpy(pRing->_wp,buf,dwSize);
    pRing->_wp+=dwSize;
  }

  pRing->frames._uAvailable-=uFrames;
  pRing->frames._uWritten+=uFrames;

  code=0;
eoverflow:
  LeaveCriticalSection(&pRing->pEngine->cs); // }
enoinit:
  return code;
}

// Dithering algorithms:
//   https://www.kvraudio.com/forum/viewtopic.php?t=132643
// Image Dithering: Eleven Algorithms and Source Code:
//   http://www.tannerhelland.com/4660/dithering-eleven-algorithms-source-code/
static BYTE *Ring2BufFrames(Ring *pRing, BYTE *pDst, const BYTE *pSrc,
    WORD uFrames)
{
	static double xSqrt2Inverse=-1.0;

  Engine *pEngine=pRing->pEngine;
  WAVEFORMATEX source=pEngine->wfx.source.Format;
  WAVEFORMATEX target=pEngine->wfx.target.Format;
  int bConvMatrix=target.nChannels!=source.nChannels;
  int bConvFormat=target.nBlockAlign!=source.nBlockAlign;
  int bApplyGain=pEngine->param.pIEngineVolumeVtbl->bApplyGain;
  int nVolume=pEngine->pPlayer->param.nVolume;
  DWORD dwSizeTarget=target.nBlockAlign*uFrames;
  BYTE *pDstMax=pDst+dwSizeTarget;
  WORD nChannel;

  sample_t sample;
  BYTE *pSampleDst=(BYTE *)&sample
      +((sizeof sample)-(source.wBitsPerSample>>3));
  BYTE *pSampleSrc=(BYTE *)&sample
      +((sizeof sample)-(target.wBitsPerSample>>3));

  bApplyGain=bApplyGain&&(0<=nVolume);
  bApplyGain=bApplyGain&&nVolume<ePlayerVolumeMax;

  if (bConvMatrix) {
    if (1u!=source.nChannels&&2u!=target.nChannels) {
      // We should never go here ...
      DVMESSAGE("matrix conversion %u channels -> %u channels not supported",
          source.nChannels,target.nChannels);
      return pDstMax;
    }

    if (bConvFormat||bApplyGain) {
      pRing->dither.lpVtbl->Start(pRing);

      while (pDst<pDstMax) {
        sample=0;
        memcpy(pSampleDst,pSrc,source.wBitsPerSample>>3);
        pSrc+=source.wBitsPerSample>>3;

        switch (pEngine->options.common.eMono2Stereo) {
        case eMono2StereoHalf:
          sample>>=1;
          break;
        case eMono2StereoSqrt2Inverse:
          if (xSqrt2Inverse<0.0)
            xSqrt2Inverse=1.0/sqrt(2.0);

          sample=xSqrt2Inverse*sample+0.5;
          break;
        default:
          break;
        }

        if (bApplyGain)
          sample=MulDiv(sample,nVolume,ePlayerVolumeMax);

        sample+=pRing->dither.lpVtbl->Noise(pRing,0);
        memcpy(pDst,pSampleSrc,target.wBitsPerSample>>3);
        pDst+=target.wBitsPerSample>>3;
        memcpy(pDst,pSampleSrc,target.wBitsPerSample>>3);
        pDst+=target.wBitsPerSample>>3;
      }
    }
    else {
      while (pDst<pDstMax) {
        memcpy(pDst,pSrc,target.wBitsPerSample>>3);
        pDst+=target.wBitsPerSample>>3;
        memcpy(pDst,pSrc,target.wBitsPerSample>>3);
        pDst+=target.wBitsPerSample>>3;

        pSrc+=source.wBitsPerSample>>3;
      }
    }
  }
  else if (bConvFormat||bApplyGain) {
    pRing->dither.lpVtbl->Start(pRing);

    while (pDst<pDstMax) {
      for (nChannel=0u;nChannel<source.nChannels;++nChannel) {
        sample=0;
        memcpy(pSampleDst,pSrc,source.wBitsPerSample>>3);
        pSrc+=source.wBitsPerSample>>3;

        if (bApplyGain)
          sample=MulDiv(sample,nVolume,ePlayerVolumeMax);

        sample+=pRing->dither.lpVtbl->Noise(pRing,nChannel);
        memcpy(pDst,pSampleSrc,target.wBitsPerSample>>3);
        pDst+=target.wBitsPerSample>>3;
      }
    }
  }
  else
    memcpy(pDst,pSrc,dwSizeTarget);

  return pDstMax;
}

int RingReadFrames(Ring *pRing, BYTE *buf, WORD uFrames)
{
  int code=-1;
  WORD nBlockAlign=pRing->pEngine->wfx.source.Format.nBlockAlign;
  DWORD dwSize=nBlockAlign*uFrames;
  DWORD dwSize1,dwSize2;

  if (!pRing->pEngine) {
    DMESSAGE("ring not initialized");
    goto enoinit;
  }

  EnterCriticalSection(&pRing->pEngine->cs); // {

  if (pRing->_rp<pRing->_wp) {
    // not wrapped
    if (pRing->_wp<pRing->_rp+dwSize) {
      DMESSAGE("underflow");
      goto eunderflow;
    }

    buf=Ring2BufFrames(pRing,buf,pRing->_rp,uFrames);
    pRing->_rp+=dwSize;
  }
  else {
    // wrapped
    dwSize1=pRing->_mp-pRing->_rp;

    if (dwSize<dwSize1) {
      buf=Ring2BufFrames(pRing,buf,pRing->_rp,uFrames);
      pRing->_rp+=dwSize;
    }
    else {
      dwSize2=dwSize-dwSize1;
      buf=Ring2BufFrames(pRing,buf,pRing->_rp,dwSize1/nBlockAlign);
      buf=Ring2BufFrames(pRing,buf,pRing->_buf,dwSize2/nBlockAlign);
      pRing->_rp=pRing->_buf+dwSize2;
    }
  }

  pRing->frames._uWritten-=uFrames;
  pRing->frames._uAvailable+=uFrames;
#if defined (YASAPI_API) // [
  SetEvent(pRing->pEngine->thread.hApi);
#endif // ]

  code=0;
eunderflow:
  LeaveCriticalSection(&pRing->pEngine->cs); // }
enoinit:
  return code;
}

/*
RingSourceSizeAvailable

RingTargetSize
RingTargetSizeWritten
*/
//DWORD RingSourceSize(Ring *pRing);
DWORD RingSourceSize(Ring *pRing)
{
  return pRing->_mp-pRing->_buf;
}

DWORD RingSourceSizeWritten(Ring *pRing)
{
  WORD nBlockAlign=pRing->pEngine->wfx.source.Format.nBlockAlign;

  return nBlockAlign*pRing->frames._uWritten;
}

DWORD RingSourceSizeAvailable(Ring *pRing)
{
  WORD nBlockAlign=pRing->pEngine->wfx.source.Format.nBlockAlign;

  return nBlockAlign*pRing->frames._uAvailable;
}

DWORD RingTargetSize(Ring *pRing)
{
  WORD nBlockAlignTarget=pRing->pEngine->wfx.target.Format.nBlockAlign;
  WORD nBlockAlignSource=pRing->pEngine->wfx.source.Format.nBlockAlign;

  return MulDiv(pRing->_mp-pRing->_buf,nBlockAlignTarget,nBlockAlignSource);
}

DWORD RingTargetSizeWritten(Ring *pRing)
{
  WORD nBlockAlign=pRing->pEngine->wfx.target.Format.nBlockAlign;

  return nBlockAlign*pRing->frames._uWritten;
}

//DWORD RingTargetSizeAvailable(Ring *pRing);
#endif // ]
