/*
 * yasapi_thread.c
 *
 * Yet Another (WA)SAPI (YASAPI) New Technology (NT) Output (out_yasapi-nt)
 * Copyright (C) 2015-2018  Peter Belkner (pbelkner@snafu.de)
 *
 * out_yasapi-nt is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * out_yasapi-nt is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with out_yasapi-nt.  If not, see <https://www.gnu.org/licenses/>.
 */
#include <yasapi.h>
#include <avrt.h>

///////////////////////////////////////////////////////////////////////////////
#if ! defined (ERROR_INVALID_TASK_NAME) // [
#define ERROR_INVALID_TASK_NAME 1550L
#endif // ]

#if ! defined (ERROR_INVALID_TASK_INDEX) // [
#define ERROR_INVALID_TASK_INDEX 1551L
#endif // ]

///////////////////////////////////////////////////////////////////////////////
typedef enum _EThread EThread;
typedef struct _Thread Thread;

///////////////////////////////////////////////////////////////////////////////
enum _EThread {
  eThreadProducer,
  eThreadTimerEvent,
  eThreadTimer,
  eThreadCount,
};

struct _Thread {
  Engine *pEngine;
  HANDLE aHandles[eThreadCount];
  // The following should be sufficient in allmost all cases. But shouldn't
  // we have a ring buffer instead to be on the save side?
  Command command;
#if defined (YASAPI_AV_THREAD_CHARACTERISTICS) // [
  HANDLE hTask;
  DWORD dwTaskIndex;
#endif // ]
};

///////////////////////////////////////////////////////////////////////////////
static void ThreadCreate(Thread *pThread, Engine *pEngine)
{
  pThread->pEngine=pEngine;
  pThread->aHandles[eThreadProducer]=pEngine->thread.hProducer;
  pThread->aHandles[eThreadTimerEvent]=pEngine->timer.hEvent;
  pThread->aHandles[eThreadTimer]=pEngine->timer.hTimer;
  pThread->command.eCommand=eCommandNull;
  pThread->command.eArgument=eEngineNull;
#if defined (YASAPI_AV_THREAD_CHARACTERISTICS) // [
  pThread->hTask=NULL;
  pThread->dwTaskIndex=0ul;
#endif // ]

  EnterCriticalSection(&pEngine->cs); // {
  ++pEngine->eState;
  DASSERT(eEngineThreadRunning==pEngine->eState,exit);
  DWRITELN(EngineState(pEngine));
  SetEvent(pEngine->thread.hConsumer);
DEXIT();
  return;
}

static void ThreadDestroy(Thread *pThread)
{
  Engine *pEngine=pThread->pEngine;

  DWRITELN(EngineState(pEngine));
  LeaveCriticalSection(&pEngine->cs); // }
  ExitThread(0ul);
}

static DWORD ThreadWait(Thread *pThread)
{
  Engine *pEngine=pThread->pEngine;
  DWORD dwWait;

  LeaveCriticalSection(&pEngine->cs); // {

  dwWait=WaitForMultipleObjects(
    eThreadCount,       // _In_       DWORD  nCount,
    pThread->aHandles,  // _In_ const HANDLE *lpHandles,
    FALSE,              // _In_       BOOL   bWaitAll,
    INFINITE            // _In_       DWORD  dwMilliseconds
  );

  EnterCriticalSection(&pEngine->cs); // }

  return dwWait;
}

static void ThreadCommandKill(Thread *pThread, const EEngine eState)
{
  Engine *pEngine=pThread->pEngine;

  EngineKill(pEngine,eState);

#if defined (YASAPI_AV_THREAD_CHARACTERISTICS) // [
  if (pEngine->eState<eEngineAudioClient&&pThread->hTask) {
    AvRevertMmThreadCharacteristics(pThread->hTask);
    pThread->hTask=NULL;
    pThread->dwTaskIndex=0ul;
  }
#endif // ]

  pThread->command.eCommand=eCommandNull;
  pThread->command.eArgument=eEngineNull;

  DWRITELN(EngineState(pEngine));
  pEngine->thread.pCommand=NULL;
  SetEvent(pEngine->thread.hConsumer);
}

static int ThreadCommandStart(Thread *pThread)
{
  int code=-1;
  Engine *pEngine=pThread->pEngine;
#if defined (YASAPI_AV_THREAD_CHARACTERISTICS) // [
  DWORD dwLastError;
#endif // ]

  if (eEngineStarted<=pEngine->eState)
    goto estart;

#if defined (YASAPI_AV_THREAD_CHARACTERISTICS) // [
  if (pEngine->options.device.bProAudio&&!pThread->hTask) {
    pThread->hTask=AvSetMmThreadCharacteristicsW(L"Pro Audio",
        &pThread->dwTaskIndex);

    if (!pThread->hTask) {
      dwLastError=GetLastError();

      switch (dwLastError) {
      DCMESSAGE("AvSetMmThreadCharacteristics",ERROR_INVALID_TASK_INDEX);
      DCMESSAGE("AvSetMmThreadCharacteristics",ERROR_INVALID_TASK_NAME);
      DCMESSAGE("AvSetMmThreadCharacteristics",ERROR_PRIVILEGE_NOT_HELD);
      DCMESSAGE("AvSetMmThreadCharacteristics",ERROR_ACCESS_DENIED);
      default:
        DVMESSAGE("AvSetMmThreadCharacteristics failed with"
            " unknow error %lu",dwLastError);
        break;
      }
    }
  }
#endif // ]

  EnginePlay(pEngine);
  code=0;
estart:
  pThread->command.eCommand=eCommandNull;
  pThread->command.eArgument=eEngineNull;

  DWRITELN(EngineState(pEngine));
  pEngine->thread.pCommand=NULL;
  SetEvent(pEngine->thread.hConsumer);

  return code;
}

static int ThreadPlay(Thread *pThread)
{
  int code=-1;
  Engine *pEngine=pThread->pEngine;
  ECommand *peCommand=&pThread->command.eCommand;
  EEngine *peArgument=&pThread->command.eArgument;
  Ring *pRing=&pEngine->ring;

  switch (pEngine->eState) {
  case eEnginePauseConnected:
  case eEnginePauseDisconnected:
    break;
  case eEngineEot:
    // We should never go here ...
    DMESSAGE("internal");
    goto exit;
  case eEngineDrainingFinal:
    ++pEngine->eState;
    DWRITELN(EngineState(pEngine));
    DASSERT_ENGINE_STATE(pEngine,eEngineEot,exit);

#if defined (YASAPI_OPTION_GAPLESS) // [
    if (pEngine->options.common.bGapless) {
      /////////////////////////////////////////////////////////////////////////
DWRITELN("KILLING ======================================================");
      EngineKill(pEngine,eEngineBase);
#if 0 // [
      memset(&pEngine->wfx.source,0,sizeof pEngine->wfx.source);
      memset(&pEngine->wfx.target,0,sizeof pEngine->wfx.target);
      memset(&pEngine->param,0,sizeof pEngine->param);
#endif // ]
      SetEvent(pEngine->thread.hEot);
    }
#if defined (YASAPI_API) // [
    else
#endif // ]
#endif // ]
#if defined (YASAPI_API) // [
      SetEvent(pEngine->thread.hApi);
#endif // ]

    break;
  case eEngineDraining:
    if (0u==RingTargetSizeWritten(pRing)) {
      ++pEngine->eState;
      DWRITELN(EngineState(pEngine));
      DASSERT_ENGINE_STATE(pEngine,eEngineDrainingFinal,exit);
      break;
    }

#if defined (YASAPI_RING_REALLOC) // [
    if (pEngine->options.common.bGapless)
      SetEvent(pEngine->thread.hEot);
#endif // ]
    // Intentional fall-through.
  default:
    if (eEngineStarted<=pEngine->eState&&EnginePlay(pEngine)<0) {
      DMESSAGE("playing");
      goto exit;
    }

    break;
  }

  if (*peCommand) {
    // Apply deferred command.
    switch (*peCommand) {
    case eCommandFlush:
      EngineFlush(pEngine);
      break;
    case eCommandKill:
      ThreadCommandKill(pThread,*peArgument);
      break;
    default:
      break;
    }

    pEngine->thread.pCommand=NULL;
    SetEvent(pEngine->thread.hConsumer);
  }

  code=0;
exit:
  return code;
}

///////////////////////////////////////////////////////////////////////////////
DWORD WINAPI EngineThreadProc(LPVOID lpParameter)
{
  Engine *pEngine=lpParameter;
  Thread thread;
  DWORD dwWait;

  ThreadCreate(&thread,lpParameter); // {

  for (;;) {
    dwWait=ThreadWait(&thread);

    switch (dwWait) {
    case WAIT_OBJECT_0+eThreadProducer:
      DASSERT(pEngine->thread.pCommand,exit);
      thread.command=*pEngine->thread.pCommand;

      switch (thread.command.eCommand) {
      case eCommandExit:
        // Will be executed immediately.
        pEngine->thread.pCommand=NULL;
        SetEvent(pEngine->thread.hConsumer);
        goto exit;
      case eCommandKill:
        // Will be executed deferred during next play in allmost all cases.
        if (eEngineThreadRunning==thread.command.eArgument) {
          // The command is issued when not playing!!!
          ThreadCommandKill(&thread,eEngineThreadRunning);
          continue;
        }

        break;
      case eCommandStart:
        // Will be executed immediately.
        if (ThreadCommandStart(&thread)<0)
          goto exit;

        continue;
      case eCommandFlush:
        // Will be executed deferred during next play.
      default:
        break;
      }

      break;
    case WAIT_OBJECT_0+eThreadTimerEvent:
    case WAIT_OBJECT_0+eThreadTimer:
      if (ThreadPlay(&thread)<0) {
        DMESSAGE("playing");
        goto exit;
      }

      break;
    /*
    case WAIT_OBJECT_0+3ul:
      if (pEngine->options.common.bGapless&&eEngineEot==pEngine->eState)
        EngineKill(pEngine,eEngineBase);

      break;
    */
    default:
      DVMESSAGE("wait error: %lu",dwWait);
      goto exit;
    }
  }
exit:
  ThreadDestroy(&thread); // }
  return 0ul;
}
